/**
 * Real-time continuous output using PortAudio.
 *
 * Reason NOT to use portaudio on Linux:
 *
 * - Fails to run if compiled to use JACK and JACK is not available
 *
 * - Does sample-rate-conversion on 44100Hz output (probably to
 *   48000Hz) even when hardware is capable of 44100Hz.  This causes
 *   aliasing noises at higher frequencies (e.g. z7.sbg).
 */

#include <portaudio.h>
#include <stdio.h>                       
#include <stdlib.h>                       
#include <string.h>

#include "api.h"

/**
 * PortAudio callback to generate more data from SBG.
 */
static int callback(
   const void *input,
   void *output,
   unsigned long frameCount,
   const PaStreamCallbackTimeInfo* timeInfo,
   PaStreamCallbackFlags statusFlags,
   void *userData
) {
   sbg_gen((SBG*) userData, (short*) output, (int) frameCount);
   return paContinue;
}

/**
 * Active audio stream, or NULL.
 */
static PaStream *stream;

extern char *sbg_pa_stop();

/**
 * Start outputting audio from the given SBG engine using PortAudio.
 * The time 'now' will start advancing immediately.  This is suited to
 * playing sequences that have already been loaded up into the engine,
 * or for playing real-time sequences which are generated in real time
 * relative to the time 'now'.  Note that 'now' will be ahead of
 * actual time according to the latency introduced by the audio
 * buffers, which depends on the platform.
 *
 * Returns NULL on success, or otherwise a strdup'd error string which
 * the caller must free.
 */
char *
sbg_pa_start(SBG *sbg) {
   PaError err;
   char *context;

   if (stream) {
      char *rv = sbg_pa_stop();
      if (rv) return rv;
   }

   err = Pa_Initialize();
   context = "PortAudio initialization";
   if (err != paNoError) goto error;

   err = Pa_OpenDefaultStream(
      &stream, 0, 2,
      paInt16, sbg_rate(sbg),
      paFramesPerBufferUnspecified,
      callback, (void*) sbg);
   context = "Audio stream open";
   if (err != paNoError) goto error;

   err = Pa_StartStream(stream);
   context = "Audio stream start";
   if (err != paNoError) goto error;

   return NULL;
   
  error:
   {
      char buf[1024];
      snprintf(buf, sizeof(buf), "%s error: %s\n", context, Pa_GetErrorText(err));
      return strdup(buf);
   }
}

/**
 * Stop playing and close the audio device.  The SBG engine will
 * freeze with whatever value 'now' had before this call.  If it is
 * restarted later, it will continue from this point.
 *
 * Returns NULL on success, or otherwise a strdup'd error string which
 * the caller must free.
 */
char *
sbg_pa_stop() {
   PaError err;
   char *context;

   if (!stream) return NULL;
   
   err = Pa_StopStream(stream);
   context = "Audio stream stop";
   if (err != paNoError) goto error;
   
   err = Pa_CloseStream(stream);
   context = "Audio stream close";
   if (err != paNoError) goto error;

   stream = NULL;

   err = Pa_Terminate();
   context = "Audio stream termination";
   if (err != paNoError) goto error;

   return NULL;

  error:
   {
      char buf[1024];
      snprintf(buf, sizeof(buf), "%s error: %s\n", context, Pa_GetErrorText(err));
      return strdup(buf);
   }
}

// END //
