//
//	Sequenced Binaural Beat Generator Library
//
//        Copyright (c) 1999-2011 Jim Peters <http://uazu.net/>.
//        Released under the GNU GPL version 2 as published by the
//        Free Software Foundation.  See the file COPYING for details,
//        or visit <http://www.gnu.org/copyleft/gpl.html>.
//
//	Note that this file is released under the *GNU GPL*, so this
//	means that it can only be linked with code that uses a
//	compatible open-source / free software licence.  For an
//	overview of compatible and incompatible licences, see here:
//
//	  http://www.fsf.org/licenses/license-list.html
//
//	If you want to use this code in a project that can't be put
//	under a compatible licence (e.g. closed-source), then you
//	need specific permission from me to use this code.
//
//	-- Jim Peters   (http://uazu.net/contact/)
//

static void
error(char *fmt, ...) {
   va_list ap; va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   fprintf(stderr, "\n");
#ifdef EXIT_KEY
   fprintf(stderr, "Press <RETURN> to continue: ");     
   fflush(stderr);                                      
   getchar();                                           
#endif                                               
   exit(1);                                          
}                                                 

static void                                              
debug(char *fmt, ...) {                           
   va_list ap; va_start(ap, fmt);                    
   vfprintf(stderr, fmt, ap);                        
   fprintf(stderr, "\n");                            
}                                                 

static void                                              
warn(char *fmt, ...) {                            
   va_list ap; va_start(ap, fmt);                    
   vfprintf(stderr, fmt, ap);                        
   fprintf(stderr, "\n");                            
}                                                 

static void *
Alloc(size_t len) {
   void *p= calloc(1, len);
   if (!p) error("Out of memory");
   return p;
}

static char *
StrDup(char *str) {
   char *rv= strdup(str);
   if (!rv) error("Out of memory");
   return rv;
}

static char *
StrDupRange(char *s, char *e) {
   int len = e-s;
   char *rv= calloc(1, len+1);
   if (!rv) error("Out of memory");
   memcpy(rv, s, len);
   rv[len] = 0;
   return rv;
}

/**
 * Integer-only cosine calculation using a 5-term Taylor expansion.
 * This is the equivalent of a 65536-entry cosine table containing
 * 16-bit values.  The amplitude of the resulting cosine is 32767.
 * Accuracy of output values is +/- 1: all 0/-1 on positive half, and
 * 0/+1 on negative half.  So this means it has about 15-bit accuracy.
 * 
 * Note that calc_cos(n) is actually equivalent to
 * 32767*cos((n+0.5)/65536*2*M_PI), i.e. the table index is offset by
 * 0.5 units, which makes the calculation easier to handle internally.
 *
 * The amplitude argument 'amp' is 0 to 65535.  amp==1 gives the
 * direct output of the cosine calculations.  Maximum 'amp' of 65535
 * gives a wave that ranges within +/- 0x7FFE8001.
 *
 * This is highly optimised.  Please don't make adjustments or
 * 'improvements' here without running checks against the expression
 * above for all values 0..65535.  This precise algorithm is quick and
 * also gives an accurate result whereas slightly different values
 * give much worse results.
 */

static inline S32
calc_cos(S16 off, U16 amp) {
   U32 v1,v2,v4,v6,v8, uu;
   S32 tt;

   off ^= off >> (sizeof(off) * 8 - 1);
   tt= off;

   v1= 32767 - tt - tt;     // Range -32767 to 32767  (-pi/2 would be -32768, pi/2 32768)
   v2= (v1*v1) >> 14;       // Range 0 to 65532 (pi^2/4 would be 65536)
   v4= (v2*v2) >> 16;       // Range 0 to 65528 (pi^4/16 would be 65536)
   v6= (v4*v2) >> 16;       // Range 0 to 65524 (pi^6/64 would be 65536)
   v8= (v4*v4) >> 16;       // Range 0 to 65520 (pi^8/256 would be 65536)

   uu= (42332*v2 - 5222*v4 + 307*v6 - 11*v8) >> 16;
   tt= (102939 - uu) * v1;
   return ((tt + 32768) >> 16) * amp;
}

/**
 * Initialise the sin table.  This does floating point calculations.
 * DO NOT CHANGE any of the constants here without rechecking the
 * harmonic response using 'noise-floor' in the testing folder.
 */

#define ST_INDEX_BITS 12
#define ST_VALUE_BITS 20
#define ST_INTERP_BITS 9

#define ST_SIZE (1<<ST_INDEX_BITS)
#define ST_AMP ((1<<(ST_VALUE_BITS-1))-1)   // e.g. 8 => -127 to 127
#define ST_INTERP_MASK ((1<<ST_INTERP_BITS)-1)

static S32 sin_table[ST_SIZE+1];
static int sin_table_init = 0;

static void
init_sin_table() {
   int a;
   if (!sin_table_init) {
      for (a = 0; a<=ST_SIZE; a++) {
         sin_table[a] = (int) floor(
            0.5 + ST_AMP * sin((a * 3.14159265358979323846 * 2) / ST_SIZE));
      }

      // largest-diff * max amplitude * max interpolation MUST NOT overflow
      S32 max_diff = sin_table[1] - sin_table[0];
      S32 max_amp = ((U32) 0xFFFFFFFF) >> ST_VALUE_BITS;
      S32 max_interp = ST_INTERP_MASK;
      if ((max_diff * max_amp * max_interp) / max_diff / max_amp / max_interp != 1) {
         fprintf(stderr, "sine table calculations will overflow: "
                 "increase ST_INDEX_BITS or decrease ST_INTERP_BITS\n");
         int ib = ST_INTERP_BITS;
         while ((max_diff * max_amp * ((1<<ib)-1)) / max_diff / max_amp / ((1<<ib)-1) != 1)
            ib--;
         fprintf(stderr, "maximum ST_INTERP_BITS == %d\n", ib);         
         error("ABORT");
      }
         
      sin_table_init = 1;
   }
}

/**
 * Lookup a sine value from a table using interpolation.  The table
 * has a 12-bit index (ST_INDEX_BITS) and contains 20-bit
 * (ST_VALUE_BITS) values.  The number of bits in the table index
 * matters a lot less than the interpolation, which brings in more
 * bits of the phase.  With an 12-bit table (ST_INDEX_BITS) and
 * interpolation, the errors drop below the noise floor when dithering
 * is applied.  This means that it gives better than 21 bits of
 * harmonic accuracy.
 *
 * Phase offset 'off' is a full-range unsigned 32-bit value,
 * expressing the range 0 to 2*PI.  The number of bits of this value
 * used in calculations is 21 (ST_INDEX_BITS + ST_INTERP_BITS).
 *
 * Amplitude is a full-range unsigned 32-bit value.  The number of
 * bits used is 12 (32-ST_VALUE_BITS).
 *
 * For an amplitude of 0xFFFFFFFF, the wave ranges within +/-
 * 0x7ff7f001.  This is calculated from maximum 0xFFF amplitude
 * multiplied by maximum 0x7FFFF from sine table.
 *
 * The harmonic qualities of this generator have been checked using
 * 'noise-floor' in the testing folder.  All harmonic errors are below
 * the 21-bit noise floor generated by the dithering.  However, making
 * any change to the ST_* constants is likely to worsen the response,
 * so recheck with 'noise-floor'.
 */
static S32
lookup_sin(U32 off, U32 amp) {
   U32 table_off = off >> (32-ST_INDEX_BITS);
   U32 interp_off = (off >> (32-ST_INDEX_BITS-ST_INTERP_BITS)) & ST_INTERP_MASK;
   S32 mult = (S32) (amp >> ST_VALUE_BITS);
   S32 v0 = sin_table[table_off] * mult;
   S32 v1 = sin_table[table_off+1] * mult;
   S32 adj = ((v1-v0) * (S32) interp_off) >> ST_INTERP_BITS;
   return v0 + adj;
}

 
// END //
