//
//	Sequenced Binaural Beat Generator Library
//
//        Copyright (c) 1999-2011 Jim Peters <http://uazu.net/>.
//        Released under the GNU GPL version 2 as published by the
//        Free Software Foundation.  See the file COPYING for details,
//        or visit <http://www.gnu.org/copyleft/gpl.html>.
//
//	Note that this file is released under the *GNU GPL*, so this
//	means that it can only be linked with code that uses a
//	compatible open-source / free software licence.  For an
//	overview of compatible and incompatible licences, see here:
//
//	  http://www.fsf.org/licenses/license-list.html
//
//	If you want to use this code in a project that can't be put
//	under a compatible licence (e.g. closed-source), then you
//	need specific permission from me to use this code.
//
//	-- Jim Peters   (http://uazu.net/contact/)
//

// Binaural beat voice.
// 
// Spec format:
//     -                            # 'off'
//     <carr>/<amp>                 # pure carrier, no beat
//     <carr><sign><beat>/<amp>     # carrier with beat
//
// <carr>: Carrier frequency in Hz, floating point number
// <beat>: Beat frequency in Hz, floating point number
// <sign>: + or -, determines which channel gets the higher frequency
// <amp>: Amplitude relative to full-scale, 0 to 100, floating point number
//

typedef struct SBGV_Bin SBGV_Bin;
struct SBGV_Bin {
   SBG_Voice v;

   S32 phaL, phaR;	// 32-bit left/right phases (2^32 <=> 2*pi)

   S32 carr;            // Carrier frequency expressed as a phase increment per sample
   S32 beat;            // Half beat frequency expressed as a phase increment per sample
   S32 carr_lo;         // 16-bit fractional part of carrier (forming a 32.16 value)
   S32 beat_lo;         // 16-bit fractional part of beat (forming a 32.16 value)
   U32 amp;             // Amplitude from 0 to 2**32-1

   S32 carr_inc;        // Increment per sample for carrier (16.16 value)
   S32 beat_inc;        // Increment per sample for beat (16.16 value)
   S32 amp_inc;         // Amplitude increment per sample
};

// Convert to S32 with limiting
static S32 to_S32(double val) {
   return val < -0x7FFFFFFF ? -0x7FFFFFFF :
      val > 0x7FFFFFFF ? 0x7FFFFFFF :
      (S32) val;
}

// Convert to U32 with limiting
static U32 to_U32(double val) {
   return val < 0 ? 0 :
      val > 0xFFFFFFFF ? 0xFFFFFFFF :
      (U32) val;
}

static SBG_Voice*
bin_init(SBG *sbg, char *spec, char **err) {
   SBGV_Bin *vv = ALLOC(SBGV_Bin);
   SBG_Voice *rv = &vv->v;
   double carr, beat, amp;
   int sign;
   Parse pp;

   *err = NULL;

   // Initialise phases to start at 0 of sin wave
   vv->phaL = 0;
   vv->phaR = 0;
   
   parse_init(&pp, spec);
   if (parse(&pp, "-%E")) {
      // Off: Leave all as 0
      return rv;
   }
   if (!parse(&pp, "%F", &carr)) {
      free(rv);
      return NULL;
   }
   beat = 0.0;
   if (parse(&pp, "%S%f", &sign, &beat)) {
      beat *= sign;
   }
   if (!parse(&pp, "/%f%E", &amp)) {
      free(rv);
      *err = StrDup("Bad binaural beat spec, expecting /<amp> at end");
      return NULL;
   }
   
   vv->carr = to_S32(floor(0.5 + (65536.0 * 65536.0) * carr / sbg->rate));
   vv->beat = to_S32(floor(0.5 + (65536.0 * 65536.0) * beat * 0.5 / sbg->rate));
   vv->amp = to_U32(floor(0.01 * 0xFFFFFFFF * amp));

   //debug("Decoded: carr %08X, beat %08X, amp %08X", vv->carr, vv->beat, vv->amp);
   
   return rv;
}

static SBG_Voice* bin_offdup(SBG *sbg, SBG_Voice *v1) {
   SBGV_Bin *vv = ALLOC(SBGV_Bin);
   memcpy(vv, v1, sizeof(*vv));
   vv->amp = 0;
   return &vv->v;
}

static void
bin_term(SBG_Voice *vv) {
   free(vv);
}

/**
 * Maximum frequency slide rate is (0x7FFF * 44100 / 2**32) * 44100 ==
 * 14800Hz per second.  Accuracy of frequency slide is (1 * 44100 /
 * 2**48) * 44100 = 0.0000069Hz per second, or 0.004Hz for a 10 minute
 * slide.
 *
 * Maximum amplitude slide rate is 0x7FFFFFFF / 2**32 == 50% per
 * sample.  Accuracy of amp slide is (1 / 2**32) * 44100 * 100% ==
 * 0.001% per second or 0.6% in a 10 minute slide.
 */

static void
bin_slide(SBG_Voice *v1, SBG_Voice *v2) {
   SBGV_Bin *b1 = (SBGV_Bin*) v1;
   SBGV_Bin *b2 = (SBGV_Bin*) v2;

   if (!b2) {
      b1->carr_inc = 0;
      b1->beat_inc = 0;
      b1->amp_inc = 0;
      return;
   }

   if (b1->v.is_off || b2->v.is_off) {
      // Fading to/from 'off', don't change frequencies
      b1->carr_inc = 0;
      b1->beat_inc = 0;
   } else {
      // Round frequency changes to nearest -- we may overshoot
      double carr_inc = floor(
         0.5 + (b2->carr - b1->carr) * 65536.0 / (b2->v.time - b1->v.time));
      double beat_inc = floor(
         0.5 + (b2->beat - b1->beat) * 65536.0 / (b2->v.time - b1->v.time));

      b1->carr_inc = to_S32(carr_inc);
      b1->beat_inc = to_S32(beat_inc);

      // Need to limit to 0x7FFF0000 because there may be a number
      // 0000-FFFF already present in the _lo variable, and we don't
      // want to overflow.
      if (b1->carr_inc > 0x7FFF0000)
         b1->carr_inc = 0x7FFF0000;
      if (b1->beat_inc > 0x7FFF0000)
         b1->beat_inc = 0x7FFF0000;
   }
   
   // Round amplitude change towards zero, to avoid overshoot
   // (which might wrap around from 0 to max amp)
   double amp_delta, amp_sign;
   if (b2->amp > b1->amp) {
      amp_delta = (double) (b2->amp - b1->amp);
      amp_sign = 1.0;
   } else {
      amp_delta = (double) (b1->amp - b2->amp);
      amp_sign = -1.0;
   }
   b1->amp_inc = to_S32((amp_sign * floor(amp_delta / (b2->v.time - b1->v.time))));

   //debug("amp %08X->%08X: amp_delta %lg, amp_sign %lg, amp_inc %08X",
   //      b1->amp, b2->amp, amp_delta, amp_sign, b1->amp_inc);
   
   //debug("Slide: carr_inc %08X, beat_inc %08X, amp_inc %08X",
   //      b1->carr_inc, b1->beat_inc, b1->amp_inc);
}
   
static void
bin_xfer(SBG_Voice *v1, SBG_Voice *v2) {
   SBGV_Bin *b1 = (SBGV_Bin*) v1;
   SBGV_Bin *b2 = (SBGV_Bin*) v2;

   b2->phaL = b1->phaL;
   b2->phaR = b1->phaR;
}

static void 
bin_gen(SBG_Voice* v1, S32 cnt, S32 *dst) {
   SBGV_Bin* vv = (SBGV_Bin*) v1;
   
   while (cnt-- > 0) {
      vv->phaL += vv->carr + vv->beat;
      vv->phaR += vv->carr - vv->beat;
      vv->carr_lo += vv->carr_inc;
      vv->carr += vv->carr_lo >> 16;
      vv->carr_lo &= 65535;
      vv->beat_lo += vv->beat_inc;
      vv->beat += vv->beat_lo >> 16;
      vv->beat_lo &= 65535;
      *dst++ += lookup_sin(vv->phaL, vv->amp);
      *dst++ += lookup_sin(vv->phaR, vv->amp);
      vv->amp += vv->amp_inc;
   }
}

/**
 * Debug dump
 */
static void
bin_dump(SBG_Voice *v1, FILE *out) {
   SBGV_Bin* vv = (SBGV_Bin*) v1;

   fprintf(out, ":\n");
   fprintf(out, "    phase L: %08X, R: %08X\n", vv->phaL, vv->phaR);
   fprintf(out, "    carrier: %08X.%04X, inc %08X\n", vv->carr, vv->carr_lo, vv->carr_inc);
   fprintf(out, "    beat:    %08X.%04X, inc %08X\n", vv->beat, vv->beat_lo, vv->beat_inc);
   fprintf(out, "    amp:     %08X,      inc %08X\n", vv->amp, vv->amp_inc);
}
   
// END //
