/**
 * Test the noise floor beneath sine waves generated via various
 * methods.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fftw3.h>
#include <math.h>
#include "sbglib.h"

int in_len;
int out_len;
double *in;
double *fft;
double *out;
int *tmp;
fftw_plan plan;

static void
init(int len) {
   {
      FILE *in = fopen(".wisdom", "r");
      if (in) {
         fftw_import_wisdom_from_file(in);
         fclose(in);
      }
   }
   
   in_len = len;
   in = fftw_malloc(in_len * sizeof(*in));
   fft = fftw_malloc((in_len + 1) * 2 * sizeof(*fft));
   plan = fftw_plan_dft_r2c_1d(in_len, in, (fftw_complex*) fft, 0);
   out = malloc((in_len + 1) * sizeof(*out));
   tmp = malloc(in_len * 2 * sizeof(*tmp));

   {
      FILE *out = fopen(".wisdom", "w");
      fftw_export_wisdom_to_file(out);
      fclose(out);
   }
}

static void 
do_fft() {
   int a;
   double *op = out;
   double *fp = fft;
   double adj = 2.0 / in_len;
   double mid = 0.5 * (in_len-1);
   double fact = M_PI / mid;
   double max = 0;

   double tot = 0.0;
   for (a = 0; a<in_len; a++) {
      double theta = (a-mid) * fact;
      double val= 0.42 + 0.5 * cos(theta) + 0.08 * cos(2.0 * theta);
      in[a] *= val;
      tot += val;
   }
   tot /= in_len;
   adj /= tot;
   
   fftw_execute(plan);
   for (a = in_len + 2; a>0; a-=2) {
      *op++ = hypot(fp[0], fp[1]) * adj;
      fp += 2;
      if (op[-1] > max) max = op[-1];
   }
   out_len = op-out;
   printf("Max: %lg\n", max);
}

static void
error(char *str) {
   fprintf(stderr, "%s\n", str);
   exit(1);
}

#define RATE 44100
#define SX 1000
#define SY 750

char disp[SX*SY];

static void clear() {
   int x, y;
   memset(disp, 128, sizeof(disp));
   for (x = 0; x<SX; x++)
      for (y = 0; y<SY; y++)
         disp[y*SX + x] = 255;
}

static int val2yy(double val) {
   val = val == 0.0 ? -9999999.0 : 20 * log10(val);
   val = (val / 144.5) + 1.0;  // 120dB == 20 bits, 144.5dB == 24 bits
   
   int yy = SY - 1 - (int)(SY * val);
   if (yy < 0) yy = 0;
   if (yy >= SY) yy = SY-1;
   return yy;
}

static void plot(char *pgm_fnam) {
   int a;

   do_fft();
   clear();

   // Plot dB levels for powers of 2
   for (a = 0; a<=24; a++) {
      double val = pow(0.5, a);
      int yy = val2yy(val);
      int xx;
      if (a % 8 == 0) {
         for (xx = 0; xx<SX; xx++) {
            if (xx % 2 == 0) continue;
            disp[yy*SX + xx] = 128;
         }
      } else {
         for (xx = 0; xx<SX; xx++) {
            if (xx % 4 != 2) continue;
            disp[yy*SX + xx] = 128;
         }
      }
   }

   // Plot dB level for each frequency
   for (a = 0; a<=SX; a++) {
      int i0 = out_len * a / SX;
      int i1 = out_len * (a+1) / SX;
      double val = 0;
      for (; i0 < i1; i0++) {
         if (out[i0] > val)
            val = out[i0];
      }
      int yy = val2yy(val);

      while (yy < SY) {
         disp[yy*SX + a] = 0;
         yy++;
      }
   }

   FILE *out = fopen(pgm_fnam, "w");
   fprintf(out, "P5 %d %d 255\n", SX, SY);
   fwrite(disp, sizeof(disp), 1, out);
   fclose(out);
}

static void
gen_sbg_16(char *spec) {
   SBG *sbg = sbg_init(RATE);
   int a;

   char *err = sbg_set(sbg, 0, 0, 0, spec);
   if (err) error(err);

   sbg_gen(sbg, (short*) tmp, in_len);
   for (a = 0; a<in_len; a++)
      in[a] = ((short*) tmp)[a*2] * (1.0 / 32768);
   
   sbg_term(sbg);
}
   
static void
gen_sbg_32(char *spec) {
   SBG *sbg = sbg_init(RATE);
   int a;

   char *err = sbg_set(sbg, 0, 0, 0, spec);
   if (err) error(err);

   sbg_gen32(sbg, tmp, in_len);
   for (a = 0; a<in_len; a++)
      in[a] = tmp[a*2] * (1.0 / 2147483648.0);
   
   sbg_term(sbg);
}

static void
gen_pure(double freq, double amp) {
   int a;
   double w = 2 * M_PI * freq / RATE;
   for (a = 0; a<in_len; a++)
      in[a] = sin(w * a) * amp;
}

static void
gen_pure_16(double freq, double amp) {
   int a;
   double w = 2 * M_PI * freq / RATE;
   const double fact = 32768.0;
   for (a = 0; a<in_len; a++)
      in[a] = floor(0.5 + fact * sin(w * a) * amp) / fact;
}

static void
gen_pure_16_dithered(double freq, double amp) {
   int a;
   double w = 2 * M_PI * freq / RATE;
   const double fact = 32768.0;
   for (a = 0; a<in_len; a++)
      in[a] = floor(drand48() + fact * sin(w * a) * amp) / fact;
}

static void
gen_lookup16_16_dithered(double freq, double amp) {
   int a;
   double w1 = 2 * M_PI / 65536;
   double w2 = 2 * M_PI * freq / RATE / w1;
   const double fact = 32768.0;
   for (a = 0; a<in_len; a++) {
      double theta = floor(w2 * a) * w1;
      in[a] = floor(drand48() + fact * sin(theta) * amp) / fact;
   }
}

// 16-bit lookup with interpolation
static void
gen_lint_16_dithered(int bits, double freq, double amp) {
   int a;
   double unit = 2 * M_PI / (1 << bits);
   double w = 2 * M_PI * freq / RATE;
   const double fact = 32768.0;
   for (a = 0; a<in_len; a++) {
      double theta = w * a;
      double th0 = floor(theta / unit) * unit;
      double th1 = th0 + unit;
      double val0 = sin(th0);
      double val1 = sin(th1);
      double val = val0 + (val1-val0) * (theta - th0) / unit;
      in[a] = floor(drand48() + fact * val * amp) / fact;
   }
}

int
main(int ac, char **av) {
   double freq;
   double amp;
   char dmy;

   ac--; av++;
   
   if (ac != 1)
      error("Expecting test frequency and amplitude (e.g. 100/10) on command-line");
   if (2 != sscanf(av[0], "%lf/%lf %c", &freq, &amp, &dmy))
      error("Can't parse frequency/amplitude");
   amp /= 100.0;
   
   //init(512 * 1024);   // about 11-12 seconds of audio @ 44100 Hz
   init(32 * 1024);   // about 0.7 seconds of audio @ 44100 Hz

   //gen_pure_16(freq, amp);
   //plot("respA_pure_16_undith.pgm");

   gen_pure_16_dithered(freq, amp);
   plot("respB_pure_16_dither.pgm");

   //gen_lint_16_dithered(11, freq, amp);
   //plot("respC_lint_11_16.pgm");
   
   //gen_lookup16_16_dithered(freq, amp);
   //plot("respC_lookup16_16.pgm");
      
   gen_sbg_16(av[0]);
   plot("respD_sbg_16.pgm");
   
   gen_pure(freq, amp);
   plot("respE_pure_64.pgm");

   gen_sbg_32(av[0]);
   plot("respF_sbg_32.pgm");

   return 0;
}
      

   
   
